/*
 *   FPG - False Positive Generator for snort
 *
 *   Author: Dirk Geschke <dirk@geschke-online.de>
 *
 *   Purpose: Uses snort rules and generates network traffic
 *            which should raise alerts in snort.
 *            Works only with snort without '-z' option...
 *            
 *            Source IP address and Destination IP address are
 *            used from the command line. So any fixed addresses 
 *            in the snort.conf are overwritten by these values.
 *            Therefore not every packet results in an alert.
 *
 *   Limitations: 
 *               - Not all ICMP packets are supported
 *               - No use of content-lists in snort.conf
 *               - fragbits seems not to work...
 *               - Destination IP should exist, otherwise we will
 *                 fail on arp-requests or at the last router
 *                 (Maybe we should add libnet_build_arp?)
 *               - newer keywords like byte_test and byte_jump
 *                 are not yet implemented
 *
 *   Requires: libnet-1.1.0 (http://www.packetfactory.net/libnet) 
 *
 */

/*
**   This program is part of the FLoP project
**
**   Copyright (C) 2002-2003 Dirk Geschke <Dirk@geschke-online.de>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

static char *rcsid="$Id: fpg.c,v 1.24 2004/07/05 20:44:53 geschke Exp $";

#include "FLoPconfig.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <netdb.h>
#include <ctype.h>
#include <libnet.h>
#include <time.h>

/* On linux AI_ALL is defined in <bind/netdb.h>. 
 * But mostly libnet uses <netdb.h>, so we should use here
 * also <netdb.h>, ignore the fuction warnings and link
 * against libbind. However we need a value for AI_ALL
 * at all. So if not defined we define it here.
 */
#ifdef HAVE_BIND_HEADER
#ifndef AI_ALL
#define AI_ALL          0x00000010
#endif
#endif

#define STRLEN 256
#define STRLARGE 4096
#define LIMIT 5

#define TH_FIN  0x01
#define TH_SYN  0x02
#define TH_RST  0x04
#define TH_PUSH 0x08
#define TH_ACK  0x10
#define TH_URG  0x20
#define TH_RES2 0x40
#define TH_RES1 0x80

#define IP_RB  0x8000
#define IP_DF  0x4000
#define IP_MF  0x2000

#ifndef HAVE_U_INT8_T 
#define u_int8_t  uint8_t
#endif
#ifndef HAVE_U_INT16_T
#define u_int16_t uint16_t
#endif
#ifndef HAVE_U_INT32_T
#define u_int32_t uint32_t
#endif

typedef struct _varEntry varEntry;

struct _varEntry 
{
  char * varName;
  char * varValue;
  varEntry * varNext;
};

typedef struct _Content ContentList;

struct _Content 
{
  char *Content;
  int depth;  /* + offset */
  int offset;
  int length;  
  ContentList * next;
};

void ReadConfig(char *filename);
void ReadVariables(char *filename);
void SetVariable(char *Name, char *Value);
void DumpVariable();
char * GetValue(char * var);
void ReplaceVariable(char *Name);
void CreateHeader(char * rtn);
inline u_int32_t GetRandom(u_int32_t Max);
int GetPort(char *Port);
void CmdLine(int argc, char *argv[]);
void Usage(char *argv);
void DecodeOTN(char * otn);
void EscapeString(char *String, char replace);
void UnEscapeString(char *String, char replace);
void StoreContent(char *content);
void StoreRPC(u_int32_t program, u_int32_t version, u_int32_t procedure);
void FreeContent(ContentList *List);
void DumpContent();
void StoreOffset(char *offset);
void StoreDistance(char *offset);
void StoreDepth(char *depth);
void StoreWithin(char *within);
void StoreUriContent(char * url);
void SetFlags(char *flags);
void SetFragBits(char *flags);
void BuildPayload();
void DumpStatistics();
void SetIsDataAt(char *value);
void SigInt();
int Convert2Bin(unsigned char *dest, char *content);
int SendPacket();

static ContentList * StartContent;
static varEntry * varFirst;
static int NrVal=0;
u_int16_t sport, dport;
char filename[STRLEN]="snort.conf";
char *path=NULL;
char *file;
char SourceAddr[STRLEN]="0.0.0.0";
char DestAddr[STRLEN]="0.0.0.0";
int SourceAddrSet=0;
int DestAddrSet=0;
char Protocol[5]="";
int verbose=0;
u_int32_t SrcIP;
u_int32_t SrcIP_org;
u_int32_t DestIP;
int Content;
int MinDataSize;
int MaxDataSize;
u_int8_t IPttl;
u_int8_t IPproto;
u_int8_t IPprotoSet;
u_int16_t IPid;
u_int16_t IPFrag;
u_int8_t ICMPType;
u_int8_t ICMPCode;
u_int16_t ICMPId;
u_int16_t ICMPSeq;
u_int8_t TCPFlags;
u_int32_t TCPSeq;
u_int32_t TCPAck;
u_int16_t TCPwin;
u_int8_t TCPchk;
u_int8_t TCPurg;
u_char * Payload=NULL;
u_int32_t PayloadSize;
u_int32_t TotalBytes;
int IgnoredRules=0;
int TotalPackets=0;
int TCPAckSet;
int TCPSeqSet;
int ICMPIdSet;
int ICMPSeqSet;
int UDPCount=0;
int TCPCount=0;
int IPCount=0;
int ICMPCount=0;
int FailedCount=0;
int RepeatCycles=1;
int Delay=0;
int TDelay=1;
int Random=0;
int endless=0;
int mlength=0; /* minimum length of a packet */
clock_t Time;
int Cycles=0;
u_int32_t MaxPackets=0;

int 
main(int argc, char *argv[])
{
#ifdef HAVE_BIND_HEADER
  int errnum;
#endif
  struct hostent *addr_ptr;

  Time=clock();
  /* First initialize 'random' seed */
  srand((int)(getpid()));
  
  /* Add signal handler to print statistics on interrupt */
  signal(SIGINT,SigInt);

  varFirst=calloc(sizeof(varEntry),sizeof(char));
  if (varFirst == NULL) 
    {
      printf("Calloc problem -> Exiting.\n");
      exit(1);
    }
  
  /* Parse command line arguments, maybe we get a new config file... */
  CmdLine(argc,argv);
  
  /*  Minimum Values we need */
  if ( (DestAddrSet==0) && (SourceAddrSet==0)) Usage(argv[0]);

  /* get ip address of source and destination */

#ifdef HAVE_BIND_HEADER
  addr_ptr=getipnodebyname(SourceAddr, AF_INET, AI_ALL, &errnum); 
#else
  addr_ptr=gethostbyname(SourceAddr);
#endif

  if (addr_ptr == NULL) 
    {
      fprintf(stderr,"%s: unknown host\n",SourceAddr);
      exit(1);
    }
  memcpy(&SrcIP_org, *(addr_ptr->h_addr_list), sizeof(u_int32_t));

#ifdef HAVE_BIND_HEADER
  freehostent(addr_ptr); 
#endif

#ifdef HAVE_BIND_HEADER
  addr_ptr=getipnodebyname(DestAddr, AF_INET, AI_ALL, &errnum); 
#else
  addr_ptr=gethostbyname(DestAddr);
#endif

  if (addr_ptr == NULL) 
    {
      fprintf(stderr,"%s: unknown host\n",DestAddr);
      exit(1);
    }

  memcpy(&DestIP,*(addr_ptr->h_addr_list), sizeof(u_int32_t));

#ifdef HAVE_BIND_HEADER
  freehostent(addr_ptr);
#endif

  if (verbose)
    {
      printf("Source IP     : %08x (%08x)\n", ntohl(SrcIP_org), SrcIP_org);
      printf("Destination IP: %08x (%08x)\n\n", ntohl(DestIP), DestIP);
    }

  /* First we have to get all variable names */
  ReadVariables(filename);

  /* We print all variable definitions */
  if (verbose) DumpVariable();
  
  /* Finally we read all rules */
  do  
    {
      ReadConfig(filename);
      if ( (MaxPackets) && (MaxPackets==TotalPackets) ) 
        {
          printf("MaxPackets reached: %u (%u)\n",MaxPackets,TotalPackets);
          break;
        } 	
      Cycles++;
    } while (endless);

  DumpStatistics();
  exit(0);
}

void 
DumpStatistics()
{
  fprintf(stderr,"\n");
  fprintf(stderr,"   oO=========================Oo\n"); 
  fprintf(stderr,"    | Dirk\'s FPG - Statistics |\n");
  fprintf(stderr,"   oO=========================Oo\n\n");
  fprintf(stderr,"   Nr of variables: %u\n",NrVal);
  fprintf(stderr,"   Repeat cycles  : %u\n",RepeatCycles);
  fprintf(stderr,"   Loop cycles    : %u\n",Cycles);
  fprintf(stderr,"   Total Cycles   : %u\n",Cycles*RepeatCycles);
  fprintf(stderr,"   Total packets  : %u\n",TotalPackets);
  fprintf(stderr,"   TCP packets    : %u\n",TCPCount);
  fprintf(stderr,"   UDP packets    : %u\n",UDPCount);
  fprintf(stderr,"   ICMP packets   : %u\n",ICMPCount);
  fprintf(stderr,"   IP packets     : %u\n",IPCount);
  fprintf(stderr,"   Ignored rules  : %u\n",IgnoredRules);
  fprintf(stderr,"   Failed packets : %u\n",FailedCount);
  fprintf(stderr,"   Bytes sent     : %u\n",TotalBytes);
  fprintf(stderr,"   Time           : %.3f seconds\n\n",
	 ((clock()-Time)*1000/CLOCKS_PER_SEC)/1000.0);
  return;
}


/* 
 * Function: ReadConfig(filename)
 * 
 * Purpose: Opens the file <filename> and reads in all entries to
 *          get a complete list of all rules.
 *
 * Author: DG
 *
 */

void 
ReadConfig(char *file)
{
  FILE *fd = NULL;
  char *token =NULL;
  char *key =NULL;
  char *var =NULL;
  char *rtn =NULL;
  char *otn =NULL;
  char line[STRLARGE];
  char rtnline[STRLARGE];
  char filename[STRLARGE];
  int i=0, len;
  static int include=0;

  if (path)
    {
      strncpy(filename,path,STRLARGE);
      strcat(filename,"/");
      strncat(filename,file,STRLARGE-strlen(path));
    }
  else
    {
      strcpy(filename,file);
    }
  if ( (fd=fopen(filename,"r")) == NULL ) 
    {
      snprintf(line,STRLARGE,"Can not open config file \"%s\"",filename);
      perror(line);
      exit(1);
    }

  if (verbose)  printf("File opened:%s \n",filename);
  rewind(fd);

  while((fgets(line, STRLARGE, fd))!=NULL ) 
    {
      i++;
      if (line[0]=='\0') continue;
      if (line[0]=='#') continue;
      /* remove newline at end. */
      len=strlen(line);
      len--;
      if (line[len]=='\n') 
	{
	  line[len]='\0';
	  len--;
	}
      /* if last character is a '\' remove it and read in the next line */
      while (line[len]=='\\')
	{
	  fgets((line+len),STRLARGE-len,fd);
	  len=strlen(line);
	  if (line[len]=='\n') 
	    {
	      line[len]='\0';
	      len--;
	    }
	}
      /* remove end of line */
      if ( (token=strchr(line,'\n')))   *token='\0';
      /* next if line empty (or only a comment) */
      var=line;
      if (var[0]==' ' || var [0]=='\t' ) var++;
      if (var[0]=='\0') continue;
      if (var[0]=='#') continue;

      /* We found an include file, so include it! */
      if (!strncasecmp(var,"include",7) )
      {
        key=strtok(var," ");
        token=strtok(NULL,"\t ");
	while ( token[0]==' ' || token[0]=='\t') token++;
        while ( strchr(token,'$')) 
	  {
	    /* We found a varible and try to replace it by the value */

	    ReplaceVariable(token);

	    /*            printf("Token now: <%s>\n",token); */
	  }
	include++;
	if (include > LIMIT) 
	  {
	    printf("Too many includes (%i)-> Exiting.\n",include);
	    exit(1);
	  }	
	/* Recursive call */
	ReadConfig(token); 
        /* Did we send maximum of packets? */
        if( (MaxPackets) && (TotalPackets==MaxPackets))
         {
            fclose(fd);
            return;
         }
	/* Okay, one include less */
	include--;
	continue;
      }

      if ( (strstr(line,"(") == NULL) || (strstr(line,")") == NULL) )
        {
	  /*           printf("\nIgnoring unknown entry in config file: \n");
		       printf("\n%i> %s\n\n",i,line); */
           continue;
        }

      rtn=line;
      while ( rtn[0]==' ' || rtn[0]=='\t') rtn++;
      if (strncasecmp(rtn,"alert",5) && strncasecmp(rtn,"log ",4)) 
	{
	  /*	  printf("Neither Alert nor Log:\n") ;
		  printf("%s\n",rtn); */
	  continue;
	}
       
      rtn = strtok(line,"(");
      otn = strtok(NULL,"\0"); 
      /* remove leading spaces or tabs */
      if ( !(token=strrchr(otn,')'))) 
	{
	  printf("Missing closing bracket in OTN part!\n");
          printf("<%s>\n",otn);
          exit(1);
        }
      *token='\0';

      /* copy rtn, if we have to replace variables it may expand to otn! */
      strcpy(rtnline,rtn);
      while ( strchr(rtnline,'$')) 
	  {
	    /* We found a varible an try to replace it by its value */
	    ReplaceVariable(rtnline);
	  }

      /* printf ("RTN> %s\n",rtnline); */
      CreateHeader(rtnline);

      if (verbose) 
	{
	  printf("-----------------------\n");
	  printf("Protocol        : %s\n",Protocol);
	  printf("Source Port     : %i\n",sport);
	  printf("Destination Port: %i\n",dport);
	}
      /*      printf ("OTN> %s\n",otn);  */
      Content=0;
      StartContent=calloc(sizeof(ContentList),sizeof(char));
      if (StartContent == NULL) 
	{
	  printf("Calloc problem -> Exiting.\n");
	  exit(1);
	}

      ICMPType=0;
      ICMPCode=0;
      ICMPId=0;
      ICMPIdSet=0;
      ICMPSeq=0;
      ICMPSeqSet=0;

      IPttl=255;
      IPproto=0;
      IPprotoSet=0;
      IPFrag=0;
      IPid=GetRandom(255);
      
      TCPFlags=TH_ACK;
      TCPSeq=0;
      TCPSeqSet=0;
      TCPAck=0;
      TCPAckSet=0;
      TCPwin=32767;
      TCPchk=0;
      TCPurg=0;

      PayloadSize=0;
      SrcIP=SrcIP_org;
      DecodeOTN(otn);

      if (verbose) DumpContent();

      mlength=0;
      BuildPayload();
      FreeContent(StartContent);
      if(SendPacket()) { printf("SendPacket() failed.\n");}
#ifdef DEBUG
      printf ("SendPackets: %u (%u)\n", TotalPackets, MaxPackets);
#endif
      /* Did we send maximum of packets? */
      if( (MaxPackets) && (TotalPackets==MaxPackets)) break;
    }
  if (verbose) printf("-----------------------\n");
  /* printf("Lines: %i\n",i); */
  fclose(fd);
}


/* 
 * Function: ReadVariables(filename)
 * 
 * Purpose: Opens the file <filename> and reads in all entries to
 *          get a complete list of all variables
 *
 * Author: DG
 *
 */

void 
ReadVariables(char *file)
{
  FILE *fd = NULL;
  char *token =NULL;
  char *key =NULL;
  char *var =NULL;
  char line[STRLARGE];
  char filename[STRLARGE];
  int i=0,len;
  static int include=0;

  if (path)
    {
      strncpy(filename,path,STRLARGE);
      strcat(filename,"/");
      strncat(filename,file,STRLARGE-strlen(path));
    }
  else
    {
      strcpy(filename,file);
    }
  if ( (fd=fopen(filename,"r")) == NULL ) 
    {
      snprintf(line,STRLARGE,"Can not open config file \"%s\"",filename);
      perror(line);
      exit(1);
    }

#ifdef DEBUG
  printf("ReadVariables opened: %i\n",fd);
#endif

  if (verbose) printf("File opened:%s \n",filename);
  rewind(fd);

  while((fgets(line, STRLARGE, fd))!=NULL ) 
    {
      i++;
      len=strlen(line);
      len--;
      if (line[len]=='\n') 
	{
	  line[len]='\0';
	  len--;
	}
      while (line[len]=='\\')
	{
	  fgets((line+len),STRLARGE-len,fd);
	  len=strlen(line);
	  if (line[len]=='\n') 
	    {
	      line[len]='\0';
	      len--;
	    }
	}
      /* remove end of line */
      if ( (token=strchr(line,'\n')) )  *token='\0';
      /* next if line empty (or only a comment) */
      if (line[0]=='\0') continue;
      if (line[0]=='#') continue;
      var=line;
      while ( var[0]==' ' || var[0]=='\t') var++;
      
      key = strtok(var," ");
      token = strtok(NULL,"\t ");
      var = strtok(NULL,"\t ");

      if (strncasecmp(key,"var",3) && strncasecmp(key,"include",6)) 
	{
	  /* printf("Neither Variable declaration nor input\n") ;
	     printf("Ignoring: <%s>\n",key);*/
	  continue;
	}

      /*      printf ("key> %s\n",key);
	      printf ("var> %s\n",token); */

      /* Variable definition found */
      if (!strncasecmp(key,"var",3) )
      {
	while ( var[0]==' ' || var[0]=='\t') var++;
	/*	printf ("val> %s\n",var); */
       
	/* Variable detected, try to replace it (if possible...) */
	if (var[0]=='$') 
	  {
	    var++; /* Skip further dollar sign */
	    var=GetValue(var); 
	  }
	
	/* register variable name and value (as a string!) */
	SetVariable(token,var);
      }

      /* We found an include file, so include it! */
      if (!strncasecmp(key,"include",6) )
      {
	/*	printf ("inlcude> %s\n",token); */
        while ( (var=strchr(token,'$')) ) 
	  {
	    /* We found a varible an try to replace it by the value */
	    ReplaceVariable(token);
            /* printf("Token now: <%s>\n",token); */
	  }
	include++;
	if (include > LIMIT) 
	  {
	    printf("Too many includes (%i)-> Exiting.\n",include);
	    exit(1);
	  }	
	/* Recursive call */
	ReadVariables(token);
	/* Okay, one include less */
	include--;
      }
    }
  /*  printf("Lines: %i\n",i); */
  fclose(fd);
#ifdef DEBUG
  printf("ReadVariables Closed: %u\n",fd);
#endif 
}


/* 
 * Function: SetVariable(Name,Value)
 * 
 * Purpose: Adds an entry for eacht new variable in a linked list.
 *          Twice set values are ignrored.
 *
 * Author: DG
 *
 */

void 
SetVariable(char *Name, char *Value)
{
  varEntry *List;

  List=varFirst;

  while ( List->varNext != NULL )
    {
      /* See if the variable already exist */
      if((List->varName != NULL) && (!strncmp(List->varName,Name,strlen(Name))))
      {
        printf("Value already set: %s: %s, not %s\n",
	       List->varName,List->varValue,Value);
	return;
      }
      /* Go to the next entry */
      List = List->varNext; 
    }

  /* Okay, we have a new variable, first allocate new memory */
  List->varNext = calloc(sizeof(varEntry),sizeof(char));
  if (List->varNext == NULL) 
    {
      printf("Calloc problem -> Exiting.\n");
      exit(1);
    }

  /* Add the new values */
  List->varName  = strdup(Name);
  List->varValue = strdup(Value);

  /* Count them, only for statistics, not really needed */
  NrVal++;

  return;
}


/* 
 * Function: DumpVariable()
 * 
 * Purpose: Dump all set variables to display
 *
 * Author: DG
 *
 */
void 
DumpVariable()
{
  varEntry *List;

  List=varFirst;
  printf("\nDumping %i variable names and values:\n",NrVal);
  if (NrVal==0) return;
  while (List->varName)
    {
      printf("<%s> = <%s>\n",List->varName,List->varValue);
      List = List->varNext;
    }
  printf("\n");
  return;
}


/* 
 * Function: GetValue(char *)
 * 
 * Purpose: Return the value of a variable named by var
 *
 * Return: Pointer to the value 
 *
 * Author: DG
 *
 */

char * 
GetValue(char * var)
{
  varEntry *List;

  List=varFirst;
  while (List->varName)
    {
      if (!strncmp(List->varName,var,strlen(var)))
	{
	  return(List->varValue);
	}
      List=List->varNext;
    }
  printf("No Variable <%s> found!\n",var);
  exit(1);
}


/* 
 * Function: ReplaceVariable(char *Name)
 * 
 * Purpose: Replaces variable in input string by the appropiate
 *          value out of the list
 *
 * Returns: The resulting string is stored in Name
 *
 * Author: DG
 *
 */

void 
ReplaceVariable(char *Name)
{
  char * varStart, * varEnd, * var;
  char varName[STRLEN];
  char varNew[STRLEN]="";
  int len;

  if (strlen(Name) > STRLEN ) 
    {
       printf("Size mismtach in ReplaceVariable!\n");
       printf("Increase STRLEN to at least %i\n",strlen(Name));
       exit(1);
    }

  /* Start of the variable */
  varStart=strchr(Name,'$');
  /* skip dollar sign */
  varStart++;

  /* Find end of variable, worst case: the whole string, so start here */
  varEnd=Name + strlen(Name);
  var=strchr(varStart,'/');
  if ( (var>varStart) && (var < varEnd) ) varEnd=var;
  var=strchr(varStart,' ');
  if ( (var>varStart) && (var < varEnd) ) varEnd=var;

  /* now extract name of variable */
  len=(int)(varEnd-varStart);
  strncpy(varName, varStart,len);
  /* end string with NULL */
  varName[len]='\0';
 
  /* First part until $ sign */
  strncpy(varNew,Name,varStart-1-Name);
  /* Replace variable by value */
  strcat(varNew,GetValue(varName));

  len=strlen(varNew)+strlen(varEnd);
  if (len >STRLEN )
    {
       printf("Size mismtach in ReplaceVariable!\n");
       printf("Increase STRLEN to at least %i\n",len);
       exit(1);
    }

  /* Last part after variable name */
  strcat(varNew,varEnd);
  /* Finally copy replaced string to the origin one */
  /* printf("varOld: <%s>\n",Name); */
  strcpy(Name,varNew);
  /* printf("varNew: <%s>\n",Name); */

  return;
}


/* 
 * Function: CreateHeader(char *rtn)
 * 
 * Purpose: Create the basic header, like source/destination port
 *          Actually we pick only one value out of a range and
 *          we ignore the direction.
 *
 * Author: DG
 *
 */

void 
CreateHeader(char * rtn)
{
  char * fac, *proto, *src, *sp, *dir, *dst, *dp;

#ifdef DEBUG
   printf("RTN: <%s>\n",rtn); 
#endif
  fac = strtok(rtn,"\t "); 
  proto=strtok(NULL,"\t ");
  src= strtok(NULL,"\t ");
  sp= strtok(NULL,"\t ");
  dir= strtok(NULL,"\t ");
  dst= strtok(NULL,"\t ");
  dp= strtok(NULL,"\t ");
#ifdef DEBUG
  printf("proto: <%s>, src: <%s>, sp: <%s>, dst: <%s>, dp: <%s>, dir: <%s>\n",
	 proto,src,sp,dst,dp,dir);
#endif
  snprintf(Protocol,5,proto);
  /* Only TCP and UDP have source/destination ports */
  /*  if ( (strncmp(proto,"tcp",3)) && (strncmp(proto,"udp",3)) ) return; */

  sport=GetPort(sp);
  dport=GetPort(dp);

  return;
}


/* 
 * Function: GetRandom(int Max)
 * 
 * Purpose: Creates a random number between 1 and Max
 *
 * Return: Random value in the range 1:Max
 *
 * Author: DG
 *
 */

inline u_int32_t 
GetRandom(u_int32_t Max)
{
  return (1+(u_int32_t) ((double) Max *rand()/(RAND_MAX+1.0)));
}


/* 
 * Function: GetPort(char *Port)
 * 
 * Purpose: Calculates an allowed port number according to the 
 *          given string. Port can be a single value or a range
 *          of values separated by a colon.
 *
 * Returns: Port number, more or less random
 *
 * Author: DG
 *
 */

int 
GetPort(char *Port)
{
  int not=0;
  int minPort=0;
  int maxPort=0;
  int resPort=0;
  char *res;

  /* If first sign is a '!' we have to omitt these ports */
  if (Port[0]=='!')
    {
      not=1;
      Port++; /* chop off exclamation mark */
    }
  /* if there is a colon in the string we have a range of ports */
  if (strchr(Port,':'))
    {
      res=strtok(Port,":");
      if (res) minPort=atoi(res);
      /* We have to take special care if the string starts with a colon */
      if (Port[0]==':') 
	{
	  maxPort=minPort;
	  minPort=1;
	}
      else
	{
	  res=strtok(NULL," ");
	  if (res) maxPort=atoi(res);
	  else maxPort=65535;
	}
    }
  else
    {
      minPort=atoi(Port);
      if (!strncmp(Port,"any",3)) minPort=GetRandom(65535);
      maxPort=minPort;
    }
  if (minPort==maxPort)
    {
      resPort=minPort;
      if (not) 
	{
	  while (resPort==minPort) resPort=GetRandom(65535);
	}
    }
  else
    {
      if (not) 
	{
	  resPort=GetRandom(65535);
	  while ( (resPort>=minPort) && (resPort<=maxPort) )
	    resPort=GetRandom(65535);
	}
      else
	{
	  resPort=GetRandom(maxPort-minPort+1);
	  resPort += minPort-1;
	}
    }
  /* A Final Check */
  if  (resPort >65535)  resPort=GetRandom(65535);
#ifdef DEBUG
  printf("Final Port: %i\n",resPort);
#endif

  return(resPort);
}

    
/* 
 * Function: CmdLine
 *
 * Purpose: Set or modify some values needed for startup
 *
 * Arguments: Commandline argv and number of elements argc
 *
 * Returns: Nothing, we set only global values
 *
 * Author: DG
 *
 */

void 
CmdLine(int argc, char* argv[]) 
{
  int opt, size;
  char options[]="hvVc:s:d:n:D:T:R:eM:";
  char *token;

  while((opt = getopt(argc, argv, options)) != -1) 
    {
      switch(opt)
	{
	case 'v':
	  printf("\n False Positive Generator for snort\n");
	  printf("\n A program which reads in a snort config file");
          printf("\n and tries to genetare false positive traffic");
	  printf("\n\n Written by Dirk Geschke\n\n");
	  printf("%s\n\n",rcsid);
	  exit(0);
	  break;
      
	case 'h':
	  Usage(argv[0]);
	  break;

	case 'e':
	  endless=1;
	  break;

	case 'M':
	  MaxPackets=atoi(optarg);
	  break;

	case 'D':
          Delay=atoi(optarg);
          if(Delay <0 ) Delay=0;
	  break;

	case 'T':
          TDelay=atoi(optarg);
          if(TDelay <0 ) TDelay=1;
          TDelay=TDelay*1024;
	  break;

	case 'R':
          Random=atoi(optarg);
          if(Random <0 ) Random=0;
          /* if (Random <1024) Random = Random * 1024;*/
	  Random = Random*1024;
	  break;

	case 'V':
	  verbose=1;
	  break;
	  
	case 'c':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  strncpy(filename,optarg, size);
          path=strdup(filename);
	  if ( (token=strrchr(path,'/')) ) 
	    {
	      *token='\0';
	      token++;
	      file=token;
#ifdef DEBUG
	      printf("full: %s\n",filename);
	      printf("Path: %s\n",path);
	      printf("name: %s\n",file);
#endif
	      strncpy(filename,file,STRLEN);
	    }
	  else
	    {
	      path=NULL;
	    }
	  filename[size]='\0';
	  break;

	case 's':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  strncpy(SourceAddr,optarg, size);
	  SourceAddr[size]='\0';
	  SourceAddrSet=1;
	  break;  

	case 'd':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  strncpy(DestAddr,optarg, size);
	  DestAddr[size]='\0';
          DestAddrSet=1;
	  break;  
	  
	case 'n':
	  RepeatCycles=atoi(optarg);
	  if (RepeatCycles==0) RepeatCycles=1;
	  break;  
	  
	default:
	  Usage(argv[0]);
	}
    }
}


/* 
 * Function: Usage
 *
 * Purpose: Display the possible command line arguments
 *
 * Arguments: argv, only needed to know the program name
 *
 * Returns: No return, simply exit...
 *
 * Author: DG
 *
 */

void 
Usage(char *argv)
{
  printf("\nUsage: %s [-hve] [-c config] [-M maxpkts] [-D count] [-R msec] [-T msec] [-n count] -s source -d destination \n",argv);
  printf("\n -c config     : configuartion file");
  printf("\n -d destination: destination ip to use");
  printf("\n -D count      : use a time delay between count packets");
  printf("\n -e            : loop endless (-M still counts!)");
  printf("\n -h            : this information");
  printf("\n -M maxpkts    : maximum packets which should be generated");
  printf("\n -n count      : how often to repeat on alert packet");
  printf("\n -s source     : source ip to use");
  printf("\n -R msec       : use a random time delay up to msec seconds between two packets");
  printf("\n -T msec       : use a msec time delay between packets, needs -D");
  printf("\n -v            : Version information");
  printf("\n\n");
  exit(0);
}


/* 
 * Function: DecodeOTN
 *
 * Purpose: Extract all relevant information to build an alerting
 *          packet
 *
 * Arguments: otn, the optional rule part.
 *
 * Returns: nothing (may be changing...)
 *
 * Author: DG
 *
 */

void 
DecodeOTN(char * otn)
{
  char * token;
  int i=0;
  char * key;
  char * value;
  char * res;
  char * resstart;
  int store=0;
  u_int32_t program, version, procedure;
 
#ifdef DEBUG
  printf("\nOTN: <%s>\n",otn); 
#endif

  MinDataSize=0; MaxDataSize=1500;
  EscapeString(otn,';');
  token=strtok(otn,"\a");
  if(token == NULL) return;
  do 
    {
      i++;
#ifdef DEBUG
      printf("token %i: %s\n",i,token);
#endif
      while ( token[0]==' ' || token[0]=='\t') token++;
      EscapeString(token,':');
      key=token;

      if (!strncmp(key,"content-list",12))
	{
	  printf("content-list not (yet) implemented\n");
	  IgnoredRules++;
	  return;
	}

      value=strchr(token,'\a');
      if (value) 
	{
	  UnEscapeString(value,';');
	  *value='\0'; 
	  value++;
	  while ( value[0]==' ' || value[0]=='\t') value++;
#ifdef DEBUG
	  printf("<%s>: <%s>\n",key,value);
#endif
	}
      /* else
       	printf("<%s>: none\n",key); */
      if ((!strncmp(key,"rpc",3)) && value)
       {
#ifdef DEBUG
          printf("rpc: <%s>\n",value);
#endif
          res=value;
          if (!strncmp(res,"*",1))
            {
              program = GetRandom(4294967295UL);
            }
          else
            {
              program=atol(res);
            }
          res = strchr(res,',')+1;
          if (!strncmp(res,"*",1))
             {
                version = GetRandom(4294967295UL);
             }
          else
             {
                version=atol(res);
             }
          res = strchr(res,',')+1;
          if (!strncmp(res,"*",1)) 
             {
               procedure = GetRandom(4294967295UL);
             }
          else
             { 
               procedure=atol(res);
             }

          StoreRPC(program, version, procedure);
	}
      if ((!strncmp(key,"msg",3)) && value)
	{
	  res=strdup(value);
	  resstart=res;
	  while (*value) 
	    {
	      if (*value=='\\') value++;
              *res=*value;
	      value++;
	      res++;
	    }
	  *res='\0';
	  printf("Rule: %s\n",resstart);
	  free(resstart);
	  continue;
	}
      if ((!strncmp(key,"content",7)) && value)
	{
	  store=1;
#ifdef DEBUG
	  printf("Content   : %s\n",value);
#endif
	  if (value[0]=='!') 
	    {
	      printf("Ignoring negative content\n");
	      store=0;
	      continue;
	    }
	  /* First and last sign should be quotes, we trust and
             remove them blindly... */
	  value[strlen(value)-1]='\0';
	  value++;
	  
	  /* Do a final check */
	  if (!value) 
	    {
	    store=0;
	    printf("Strange content value: Ignoring\n");
	    continue;
	  }
	  Content++;
	  StoreContent(value);
	}
      if ((!strncmp(key,"uricontent",10)) && value)
	{
	  store=1;
	  if (value[0]=='!') 
	    {
	      printf("Ignoring negative content\n");
	      store=0;
	      continue;
	    }
	  /* First and last sign should be quotes, we trust and
             remove them blindly... */
	  value[strlen(value)-1]='\0';
	  value++;
	  
	  /* Do a final check */
	  if (!value) 
	    {
	    store=0;
	    printf("Strange content value: Ignoring\n");
	    continue;
	  }
	  StoreUriContent(value);
	}
      if ( (!strncmp(key,"dsize",5)) && value )
	{
	  if (value[0]=='<') MaxDataSize=atoi(&value[1])-1;
	  else if (value[0]=='>') MinDataSize=atoi(&value[1])+1;
	  else 
	    { 
	      MinDataSize=atoi(value);
	      MaxDataSize=MinDataSize;
	    }
#ifdef DEBUG
          printf("Dsize: %s\n",value);
	  printf("DataSize between %i and %i\n",MinDataSize,MaxDataSize);
#endif
	}
      if ( (!strncmp(key,"sameip",6)) )
	{
	  SrcIP=DestIP;
	}
      if ( (!strncmp(key,"offset",6)) && store && value)
	{
#ifdef DEBUG
	  printf("Offset    : %s\n",value);
#endif
	  StoreOffset(value);
	}
/* 
 * distance and within are a little bit different to offset and depth 
 * these values are counted from the last match...
 */
      if ( (!strncmp(key,"distance",8)) && store && value)
	{
#ifdef DEBUG
	  printf("Distance  : %s\n",value);
#endif
	  StoreDistance(value);
	}
      if ((!strncmp(key,"depth",5)) && store && value)
	{
#ifdef DEBUG
	  if (value) printf("Depth     : %s\n",value);
#endif
	  StoreDepth(value);
	}
      if ((!strncmp(key,"within",6)) && store && value)
	{
#ifdef DEBUG
	  if (value) printf("Within     : %s\n",value);
#endif
	  StoreWithin(value);
	}
      if (!strncmp(key,"tos",3)) 
	printf("tos not (yet) implemented\n");
      if (!strncmp(key,"ipoption",8))
	printf("ipoption not (yet) implemented\n");
      if (!strncmp(key,"isdataat",8))
        {
#ifdef DEBUG
           printf("isdataat: <%s>\n",value);
#endif
           SetIsDataAt(value);
        }
      if (!strncmp(key,"fragbits",8) && value )
	{
	  SetFragBits(value);
	}
      if (!strncmp(key,"id",2) && value) 
	{
	  IPid=atoi(value);
	}
      if (!strncmp(key,"ip_proto",8) && value)
	{
	  IPprotoSet=1;
	  if (value[0]=='>') 
	    {
	      value++;
	      IPproto=GetRandom(255-atoi(value))+atoi(value);
	    }
	  else if (value[0]=='<')
	    {
	      value++;
	      IPttl=GetRandom(atoi(value));
	    }
	  else if (value[0]=='!') 
	    {
	      value++;
	      IPproto=GetRandom(255);
	      while (IPproto == atoi(value)) IPproto=GetRandom(255);
	    }
	  else IPproto=atoi(value);
	}
      if (!strncmp(key,"ttl",3) && value )
	{
	  if (value[0]=='>') 
	    {
	      value++;
	      IPttl=GetRandom(255-atoi(value))+atoi(value);
	    }
	  else if (value[0]=='<')
	    {
	      value++;
	      IPttl=GetRandom(atoi(value));
	    }
	  else IPttl=atoi(value);
	}
      /* ICMP specific options */
      if (!strncmp(Protocol,"icmp",4))
	{
	  if ( (!strncmp(key,"itype",5)) && value ) 
	    {
	      ICMPType=atoi(value);
	    }
	  if ( (!strncmp(key,"icode",5)) && value ) 
	    {
	      ICMPCode=atoi(value);
	    }
	  if ( (!strncmp(key,"icmp_id",7)) && value ) 
	    {
	      ICMPId=atoi(value);
	      ICMPIdSet=1;
	    }
	  if ( (!strncmp(key,"icmp_seq",8)) && value ) 
	    {
	      ICMPSeq=atoi(value);
	      ICMPSeqSet=1;
	    }
	}
      /* Only TCP relevant stuff */
      if (!strncmp(Protocol,"tcp",3))
	{
	  if ( (!strncmp(key,"flags",5)) && value ) 
	    {
	      TCPFlags=0;
	      SetFlags(value);
	    }
	  if ( (!strncmp(key,"flow",4)) && value )
	    {
	      /* All we can do here is set the ACK bit */
	      TCPFlags |=  TH_ACK;
	    }
	  if ( (!strncmp(key,"seq",3)) && value)
	    {
	      TCPSeq=atoi(value);
	      TCPSeqSet=1;
#ifdef DEBUG
	      printf("Sequence  : %s, %u\n",value,TCPSeq);
#endif
	    }
	  if ( (!strncmp(key,"ack",3)) && value )
	    {
	       TCPAck=atoi(value);
	       TCPAckSet=1;
#ifdef DEBUG
	       printf("Acknowl.  : %s, %u\n",value,TCPAck);
#endif
	    }
	}
    } while ( (token=strtok(NULL,"\a")) );
#ifdef DEBUG
  printf("--------------------------------------------\n");
#endif
  return;
}


/* 
 * Function: EscapeString
 *
 * Purpose: Replaces all escaped values of kind 'replace' by
 *          another character which should not occur in the 
 *          input. This is needed to use strtok without errors.
 *          for example: EscapeString(Input,':') results in
 *          Input : "ab\:cd:ef"
 *          Output: "ab\:cd\aef"
 *
 * Arguments: String with escape characters and character which
 *            should be removed in the escape format.
 *
 * Author: DG
 *
 */

void 
EscapeString(char *String,char replace)
{
  char *token;

#ifdef DEBUG 
  char *Start;
  Start =String;
  printf("replace: <%c>\n",replace);
  printf("Input  : <%s>\n",Start);
#endif

  while ( (token=strchr(String,replace)) )
    {
      if (token[-1]=='\\') 
        {
          String=token+1;
          continue;
        }
      *token='\a';
      String=token+1;
    }
#ifdef DEBUG
  printf("Output : <%s>\n",Start);
#endif
}


/* 
 * Function: UnEscapeString
 *
 * Purpose: Replaces all escaped values of EscapeString by
 *          its origin character
 *          for example: UnEscapeString(Input,':') results in
 *          Input  : "ab\\acd:ef"
 *          Output : "ab\:cd:ef"
 *
 * Arguments: String with escape characters and character which
 *            should be placed back
 *
 * Author: DG
 *
 */

void 
UnEscapeString(char *String,char replace)
{
  char *token;
  
  while ( (token=strchr(String,'\a')))
    {
      *token=replace;
    }
}


/* 
 * Function: FreeContent
 *
 * Purpose: Frees all allocated memory to store the contents of
 *          an alert. Since these values are stored in a linked
 *          list we free the memory recursive.
 *
 * Arguments: pointer to the list which should be freed
 *
 * Author: DG
 *
 */

void 
FreeContent(ContentList *List)
{
  if (List->next) 
    {
      FreeContent(List->next);
    }
  free(List->Content);
  free(List);
}


/* 
 * Function: StoreContent
 *
 * Purpose: Allocates memory and stores the content string in
 *          it. Additionally there are two variables for offset
 *          and depth where the final content should be stored
 *
 * Arguments: pointer to the start of the content
 *
 * Author: DG
 *
 */

void 
StoreContent(char *content)
{
  ContentList *List;
  char *dest;
  char *ptr;
  int len, escape;

  List=StartContent;
  escape=0;
  
  /* find end of list */
  while ( List->next != NULL )
    {
      List = List->next; 
    }

  /* Okay, we have a new variable, first allocate new memory 
     for the list entry */

  List->next = calloc(sizeof(ContentList),sizeof(char));
  if (List->next == NULL) 
    {
      printf("Calloc problem -> Exiting.\n");
      exit(1);
    }

  /* Add the new value: First allocate memory for the whole contents
     plus one character for the end of the string ('\0') */
  List->offset=-1;

  List->Content = malloc(strlen(content)+1);
  if (List->Content == NULL) 
    {
      printf("Calloc problem -> Exiting.\n");
      exit(1);
    }

  dest=(char *) List->Content;

  /* Now copy the content to the dest*/
  while (*content)
    {
      if (*content=='\\') 
	{
	  content++;
          escape=1;
	}
      *dest=*content;
      if ((*dest=='|') && (!escape)) 
	{
	  content++;
	  ptr=strchr(content,'|');
	  len=Convert2Bin(dest,content);
	  dest +=len;
	  content=ptr+1;
	}
      else
	{
	  dest++;
	  content++;
	}
      escape=0;
    }
  *dest='\0';
  /* Calculate the length of the content and store it. 
     Note: strlen will no longer work correctly since 
     from now on '\0' are allowed signs... */

  List->length = dest - List->Content;

#ifdef DEBUG
  printf("Len: %i\n",List->length);
#endif

  return;
}

void 
StoreRPC(u_int32_t program, u_int32_t version, u_int32_t procedure)
{
  u_int32_t flength,xid, rpcvers, msgtype;
  u_int32_t * data;
  u_int32_t * start;
  ContentList *List;
  
  flength=40;
  msgtype=0;
  rpcvers=2;
  xid=GetRandom(4294967295UL);


#ifdef DEBUG
  printf("Program  : %u\n",program);
  printf("Version  : %u\n",version);  
  printf("Procedure: %u\n",procedure);
#endif

  /* Now build the RPC data */
  start=malloc(flength);
  data=start;

  /* TCP starts with flags and length, UDP starts with transcation id */
  if (!strncmp(Protocol,"tcp",3))
    {
       *data=htonl(flength);
        data++;
    }
  *data=htonl(xid);
  data++;
  *data=htonl(msgtype);
  data++;
  *data=htonl(rpcvers);
  data++;
  *data=htonl(program);
  data++;
  *data=htonl(version);
  data++;
  *data=htonl(procedure);

  /* Now we have the data and can store them in the packet */
  List=StartContent;
  /* find end of list */
  while ( List->next != NULL )
    {
      List = List->next; 
    }

  /* Okay, we have a new variable, first allocate new memory 
     for the list entry */

  List->next = calloc(sizeof(ContentList),sizeof(char));
  if (List->next == NULL) 
    {
      printf("Calloc problem -> Exiting.\n");
      exit(1);
    }

  /* Now add the data */
  List->Content=(char *)start;
  List->length=flength;
  List->offset=0;

#ifdef DEBUG
  printf("Len: %i\n",List->length);
#endif
  return;
}


/* 
 * Function: DumpContent
 *
 * Purpose: Print all contents of the linked list.
 *          Only useful for debugging...
 *
 * Author: DG
 *
 */

void 
DumpContent()
{
  ContentList *List;
  int i=0,j;
  char c;

  List=StartContent;
  if (List->Content)  printf("========================================\n");
  
  /* find end of list */
  while ( List->next != NULL )
    {
      if (List->Content) 
	{
	  i++;
	  printf("Content: >");
	  for (j=0; j<List->length;j++)
	    {
	      c=List->Content[j];
	      if (isprint((int)c))
		putchar((int)c);
	      else 
		printf(".");  /* not printable */
	      };
	  printf("<\n");
	  printf("Offset : %i\n",List->offset);
	  printf("Depth  : %i\n",List->depth);
	  printf("Length : %i\n",List->length);
	}
      List = List->next; 
    }
  if (List->Content) 
    {
      i++;
      printf("Content: >");
      for (j=0; j<List->length;j++)
	{
          c=List->Content[j];
	  if (isprint((int)c))
	    putchar((int)c);
	  else 
	    printf("."); /* not printable */
	  c++;
	};
      printf("<\n");
      printf("Offset : %i\n",List->offset);
      printf("Depth  : %i\n",List->depth);
      printf("Length : %i\n",List->length);
    }
  if (i) printf("========================================\n");

  return;
}


/* 
 * Function: StoreOffset
 *
 * Purpose: stores the offset for the content
 *
 * Argument: value to store (as a character string!)
 *
 * Author: DG
 *
 */

void 
StoreOffset(char *offset)
{
  ContentList *List;
  ContentList *LastList;

  List=StartContent;
  LastList=List;
  
  while ( List->next != NULL )
    {
      LastList=List;
      List=List->next;
    }
  LastList->offset=atoi(offset);
  return;
}


/* 
 * Function: StoreDistance
 *
 * Purpose: stores the distance (similar to offset) for the content
 *          the distance is counted from the end of the last match
 *          hence the content length of the previous entry
 *
 * Argument: value to store (as a character string!)
 *
 * Author: DG
 *
 */

void 
StoreDistance(char *distance)
{
  ContentList *List;
  ContentList *LastList;
  int off;
  int len;

  off=0;
  len=0;
  List=StartContent;
  LastList=List;
  
  while ( List->next != NULL )
    {
      LastList=List;
      if (List->offset==-1)
       {
         List->offset=0;
       }
      else
       {
          len=List->length;
          off=List->offset;
        }
      List=List->next;
    }
  LastList->offset=atoi(distance)+off+len;
#ifdef DEBUG
  printf("distance: %i\n",LastList->offset);
#endif
  return;
}


/* 
 * Function: StoreDepth
 *
 * Purpose: stores the depth for the content
 *
 * Argument: value to store (as a character string!)
 *
 * Author: DG
 *
 */

void 
StoreDepth(char *depth)
{
  ContentList *List;
  ContentList *LastList;

  List=StartContent;
  LastList=List;
  
  while ( List->next != NULL )
    {
      LastList=List;
      List=List->next;
    }
  LastList->depth=atoi(depth);
  return;
}


/* 
 * Function: StoreWithin
 *
 * Purpose: stores the within (similiar to depth) for the content
 *
 * Argument: value to store (as a character string!)
 *
 * Author: DG
 *
 */

void 
StoreWithin(char *within)
{
  ContentList *List;
  ContentList *LastList;
  int len,off;

  len=0;
  off=0;
  List=StartContent;
  LastList=List;
  
  while ( List->next != NULL )
    {
      LastList=List;
      /* len of last content */
      len=List->length;
      off=List->offset;
      List=List->next;
    }
  LastList->depth=atoi(within)+len+off;
  return;
}


/* 
 * Function: StoreUriContent
 *
 * Purpose: stores the Uricontent. Therefore the content
 *          is extended to look like a URL request and
 *          then StoreContent is called
 *
 * Argument: value to store (as a character string!)
 *
 * Author: DG
 *
 */

void 
StoreUriContent(char * url)
{
  char *content;
  int size;
  
  size=strlen(url)+strlen(DestAddr)+22;
  content=malloc(strlen(url)+STRLEN);
  /* Build an uri request */
  snprintf(content,size,"GET http://%s/%s HTTP/1.0",DestAddr,url);
  /*  printf("UriContent now: <%s>",content); */
  /* and call StoreContent */
  StoreContent(content);
  free(content);
}


/* 
 * Function: SetFlags
 *
 * Purpose: analyzes flags statement and translates it
 *          to the appropiate value for the tcp header
 *
 * Argument: flags to store (as a character string!)
 *
 * Author: DG
 *
 */

void 
SetFlags(char *flags)
{
  int i;
  int len;
  
  len=strlen(flags);

#ifdef DEBUG
  printf("Flags: %s - ",flags);
#endif

  for (i=0; i<len; i++) 
    {
      if ( (flags[i] == '+' )  || (flags[i] == '*') )
	{
          /* ignore minimum settings flag, we alway set the minimum... */
	  continue;
	}
      if (flags[i] == '!' )
	{
	  /* Flags should not be set */
	  i++;
	  continue;
	}
      if (flags[i] == 'S' ) 
	{
	  TCPFlags |= TH_SYN;
	  continue;
	}
      if (flags[i] == 'A')
	{
	  TCPFlags |= TH_ACK;
	  continue;
	}
      if (flags[i] == 'F')
	{
	  TCPFlags |= TH_FIN;
	  continue;
	}
      if (flags[i] == 'R')
	{
	  TCPFlags |= TH_RST;
	  continue;
	}
      if (flags[i] == 'P')
	{
	  TCPFlags |= TH_PUSH;
	  continue;
	}
      if (flags[i] == 'U')
	{
	  TCPFlags |= TH_URG;
	  continue;
	}
      if (flags[i] == '1')
	{
	  TCPFlags |= TH_RES1;
	  continue;
	}
      if (flags[i] == '2')
	{
	  TCPFlags |= TH_RES2;
	  continue;
	}
      if (flags[i] == '0')
	{
	  TCPFlags = 0;
	  break;
	}
    }
#ifdef DEBUG
      printf("%02x\n",TCPFlags);
#endif
      return;
}


/* 
 * Function: SetFlags
 *
 * Purpose: analyzes flags statement and translates it
 *          to the appropiate value for the tcp header
 *
 * Argument: flags to store (as a character string!)
 *
 * Author: DG
 *
 */

void 
SetFragBits(char *flags)
{
  int i;
  int len;
  
  len=strlen(flags);

#ifdef DEBUG
  printf("Flags: %s - ",flags);
#endif

  for (i=0; i<len; i++) 
    {
      if ( (flags[i] == '+' )  || (flags[i] == '*') )
	{
          /* ignore minimum settings flag, we alway set the minimum... */
	  continue;
	}
      if (flags[i] == '!' )
	{
	  /* Flags should not be set */
	  i++;
	  continue;
	}
      if (flags[i] == 'R' ) 
	{
	  IPFrag |= IP_RB;
	  continue;
	}
      if (flags[i] == 'D')
	{
	  IPFrag |= IP_DF;
	  continue;
	}
      if (flags[i] == 'M')
	{
	  IPFrag |= IP_MF;
	  continue;
	}
    }
#ifdef DEBUG
      printf("%02x\n",IPFrag);
#endif
      return;
}


/* 
 * Function: BuildPayload
 *
 * Purpose: builds the payload for the packet
 *
 * Argument: none (works on global values)
 *
 * Author: DG
 *
 */

void 
BuildPayload()
{
  ContentList *List;
  int i=0;
  u_char * ptr=NULL;
  u_char * last=NULL;
  
  /* We need only once to allocate memory */
  if (Payload==NULL) Payload=malloc(1460);
  if (Payload==NULL)
    {
      printf("Problem to allocate memory for payload -> Exiting.\n");
      exit(1);
    }
  PayloadSize=0;

  List=StartContent;
  if (! List->Content)  return; /* no content, so return */
  
  /* find end of list */
   while ( List->next != NULL )
    {
      if (List->Content) 
	{
	  i++;
#ifdef DEBUG
          printf("List offset : %i\n",List->offset);
#endif
          if ( (List->offset+List->length) > mlength) 
	    {
	      mlength=(List->offset+List->length);
	    }
	  if (List->offset==-1) List->offset=0;
	  if (List->offset==0)
	    {
	      ptr = last;
	    }
	  if ( List->depth && 
	       ( (int)(ptr-Payload) > List->depth ) )
            {
	      printf("Oops\n");
#ifdef DEBUG
              printf("List Content: %s\n",List->Content);
              printf("List length : %i\n",List->length);
              printf("List depth  : %i\n",List->depth);
              printf("List offset : %i\n",List->offset);
              printf("ptr-Payload : %i\n",(int)(ptr-Payload));
#endif
             }
	  
	  ptr = List->offset + Payload;
	  memcpy(ptr, List->Content, List->length);
	  PayloadSize += List->length;
	  last=ptr+List->length;
	  last++;
	}
      List = List->next; 
    }
  if (List->Content) 
    {
      i++;
#ifdef DEBUG
      printf("List offset : %i\n",List->offset);
#endif
      if (List->offset==0)
	ptr = last;
      if ( List->depth && ( (int)(ptr-Payload) > List->depth ) )
	printf("Oops\n");
      ptr=List->offset+Payload;
      memcpy(ptr, List->Content, List->length);
      PayloadSize += List->length;
    }
#ifdef DEBUG
  printf("Stored length: %u\n",PayloadSize);
#endif
  if (PayloadSize < mlength) PayloadSize=mlength;

  return;
}


/* 
 * Function: Convert2Bin
 *
 * Purpose: Converts binary data (represented in hex) to real
 *          binary data
 *
 * Argument: pointer to the destianation of the data and the source
 *
 * Returns: count of stored bytes, needed to calculate end pointer
 *
 * Author: DG
 *
 */

int 
Convert2Bin(unsigned char *dest, char *content)
{
  char inp[3];
  char *ptr;
  int len;
  int i,j;
  char inpstring[STRLARGE];

  /* find end of binary part, this can't be an escaped pipe! */
  ptr=strchr(content,'|');
  /* cut string */
  *ptr='\0';
  /* ptr to next start of pointere*/
  ptr++;

  /* Remove spaces in input string */

  len=strlen(content);
  for (i=0,j=0; i<len; i++) 
    {
      if (content[i]!=' ') 
	{
	  inpstring[j]=content[i];
	  j++;
	}
    }
  inpstring[j]='\0';

#ifdef DEBUG
  printf("before  <%s>\n",content);
  printf("after   <%s>\n",inpstring);
#endif

  len=strlen(inpstring);
  if (len%2) 
    {
      printf("WARNING: size mismatch!\n");
    }
  for (i=0,j=0; i <len; i+=2,j++) 
    {
      snprintf(inp,3,"%s",&inpstring[i]);
      dest[j]= (char ) strtol(inp,NULL,16);
#ifdef DEBUG
      printf("dest: %d\n",dest[j]);
#endif
    }

#ifdef DEBUG
  printf("out: <%s>\n",dest);
#endif

  return(j);
}


/* 
 * Function: SendPacket
 *
 * Purpose: Build the final network packet and
 *          send it to the wire
 *
 * Author: DG
 *
 */

int 
SendPacket()
{
  int i,c;
  libnet_t *l;
  libnet_ptag_t t;
  int HavePacket;
  char errbuf[LIBNET_ERRBUF_SIZE];
  u_int16_t IPlen;

  HavePacket=0;

  /* Check if the dsize value can be satisfied and is useable */
  if (MinDataSize >1460) 
    {
      printf("Dsize %i too large, ignoring rule\n",MinDataSize);
      IgnoredRules++;
      return(1);
    }

  if (PayloadSize >MaxDataSize) 
    {
      printf("Dsize %i too small, ignoring rule\n",MaxDataSize);
      IgnoredRules++;
      return(1);
    }

  if (PayloadSize < MinDataSize ) PayloadSize=MinDataSize;

  /* initialize libnet */

  l = libnet_init(LIBNET_RAW4, NULL, errbuf);  
  if (l == NULL)
    {
      fprintf(stderr, "libnet_init() failed: %s", errbuf);
      exit(1); 
    }

  /**************/
  /* TCP packet */
  /**************/

  if (!strncmp(Protocol,"tcp",3))
    {
      if (!TCPSeqSet) TCPSeq=GetRandom(4294967295ul);
      if (!TCPAckSet) TCPAck=GetRandom(4294967295ul);

      if (PayloadSize==0) Payload=0;

#ifdef DEBUG
      printf("Seq: %u\n",TCPSeq);
      printf("Ack: %u\n",TCPAck);
      printf("Flg: %x\n",TCPFlags);
      printf("win: %x\n",TCPwin);
      printf("Pay: %u\n",PayloadSize);
#endif

      /* build tcp packet */

      t = libnet_build_tcp(sport, dport, TCPSeq, TCPAck, TCPFlags,
			   TCPwin, TCPchk, TCPurg,
			   LIBNET_TCP_H + PayloadSize,
			   Payload, PayloadSize, l, 0);
      
      if (t == -1)
	{
	  fprintf(stderr, "Can't build TCP header: %s\n",libnet_geterror(l));
	  libnet_destroy(l);
	  return (EXIT_FAILURE);
	}

#ifdef DEBUG
      printf("Src: %08x, Dst: %08x\n",SrcIP,DestIP);
      printf("TTL: %u, Id: %u\n",IPttl,IPid);
      printf("Total length %u\n",LIBNET_IPV4_H+LIBNET_TCP_H+PayloadSize);
#endif

      /* ... and finally build the IP packet */

      t = libnet_build_ipv4( LIBNET_IPV4_H + LIBNET_TCP_H + PayloadSize, 
			     0, IPid, IPFrag, IPttl, IPPROTO_TCP, 0, 
			     SrcIP,  DestIP, NULL, 0, l, 0);
      if (t == -1)
	{
	  fprintf(stderr, "Can't build IP header: %s\n", libnet_geterror(l));
	  libnet_destroy(l);
	  return (EXIT_FAILURE);
	}
      
      /* We have a valid packet which can be send. */
      HavePacket=1;

      /* Count the packet */
      TCPCount++;
    }

  /**************/
  /* UDP packet */
  /**************/

  if (!strncmp(Protocol,"udp",3))
    {
      if (PayloadSize==0) Payload=0;
      t=libnet_build_udp(sport,dport,LIBNET_UDP_H + PayloadSize,0, Payload,
			 PayloadSize, l, 0);
#ifdef DEBUG
      printf("UDP:\n");
      printf("PayloadSize: %i\n",PayloadSize);
      printf("Libnet_UDP_H: %i\n",LIBNET_UDP_H);
      printf("Payload     : %x\n",Payload);
#endif

      if (t == -1)
	{
	  fprintf(stderr, "Can't build UDP header: %s\n",libnet_geterror(l));
	  libnet_destroy(l);
	  return (EXIT_FAILURE);
	}
#ifdef DEBUG
      printf("UDP: id: %u, ttl: %u",IPid,IPttl);
#endif

      t = libnet_build_ipv4( LIBNET_IPV4_H + LIBNET_UDP_H + PayloadSize, 0,
			     IPid, IPFrag, IPttl, IPPROTO_UDP, 0, SrcIP,DestIP,
			     NULL, 0, l, 0);
      if (t == -1)
	{
	  fprintf(stderr, "Can't build IP header: %s\n", libnet_geterror(l));
	  libnet_destroy(l);
	  return (EXIT_FAILURE);
	}

      HavePacket=1;

      UDPCount++;      
    }

  /*************/
  /* IP packet */
  /*************/

  if (!strncmp(Protocol,"ip",2))
    {
      if (!IPprotoSet) IPproto=GetRandom(255);
#ifdef DEBUG
      printf("IP-Protocol: 0x%02x\n",IPproto);
      printf("IP-Fragment: 0x%x\n",IPFrag);
      printf("IP-Id      : 0x%x\n",IPid);
#endif
      /* add a random value for fragoffset */
      IPFrag |= GetRandom(255);

      if (PayloadSize==0) Payload=0;
      t = libnet_build_ipv4( LIBNET_IPV4_H + PayloadSize,0,
			     IPid, IPFrag, IPttl, IPproto, 0, SrcIP,  DestIP,
			     Payload, PayloadSize, l, 0);
      if (t == -1)
	{
	  fprintf(stderr, "Can't build IP header: %s\n", libnet_geterror(l));
	  libnet_destroy(l);
	  return (EXIT_FAILURE);
	}
      HavePacket=1;
      IPCount++;      
    }

  /***************/
  /* ICMP packet */
  /***************/

  if (!strncmp(Protocol,"icmp",4))
    {
      if (!ICMPIdSet) ICMPId=GetRandom(65535);
      if (!ICMPSeqSet) ICMPSeq=GetRandom(65535);

      switch (ICMPType)
	{

            /* echo request/reply */
	case 0:
	case 8: 
	  t=libnet_build_icmpv4_echo(ICMPType,ICMPCode,0,ICMPId, ICMPSeq,
				     Payload, PayloadSize, l ,0);
	  if (t == -1)
	    {
	      fprintf(stderr, "Can't build ICMP header: %s\n", 
		      libnet_geterror(l));
	      libnet_destroy(l);
	      return (EXIT_FAILURE);
	    }
	  t = libnet_build_ipv4(LIBNET_IPV4_H + LIBNET_ICMPV4_ECHO_H 
				+ PayloadSize, 0, IPid, IPFrag, IPttl,
				IPPROTO_ICMP, 0, SrcIP, DestIP, NULL, 0, l, 0);
	  if (t == -1)
	    {
	      fprintf(stderr, "Can't build IP header: %s\n", libnet_geterror(l));
	      libnet_destroy(l);
	      return (EXIT_FAILURE);
	    }
	  HavePacket=1;
	  break;

                      /* unreachable */
	case 3: 
	  IPlen=GetRandom(65535);

#ifdef HAVE_LIBNET_LIBNET_TYPES_H 
	  t = libnet_build_icmpv4_unreach ( ICMPType, ICMPCode, 0, 
					    Payload, 
					    PayloadSize, l ,0);
#else
          t = libnet_build_icmpv4_unreach ( ICMPType, ICMPCode, 0, IPlen, 0, 
                                            IPid, IPFrag, IPttl, IPproto, 0, 
                                            DestIP, SrcIP, Payload, 
                                            PayloadSize, l ,0);
#endif
	  if (t == -1)
	    {
	      fprintf(stderr, "Can't build ICMP header: %s\n", 
		      libnet_geterror(l));
	      libnet_destroy(l);
	      return (EXIT_FAILURE);
	    }
	  t = libnet_build_ipv4(LIBNET_IPV4_H + LIBNET_ICMPV4_UNREACH_H 
				+ PayloadSize, 0, IPid, IPFrag, IPttl,
				IPPROTO_ICMP, 0, SrcIP, DestIP, NULL, 0,l,0);
	  if (t == -1)
	    {
	      fprintf(stderr, "Can't build IP header: %s\n", 
		      libnet_geterror(l));
	      libnet_destroy(l);
	      return (EXIT_FAILURE);
	    }
	  HavePacket=1;
	  break;

	case 9:  /* router advertisement */
        case 10: /* router solicitation */
	  IPlen=GetRandom(65535);
	  
#ifdef HAVE_LIBNET_LIBNET_TYPES_H
	  t=libnet_build_icmpv4_redirect(ICMPType, ICMPCode, 0, DestIP, 
					 Payload, 
					 PayloadSize, l ,0);
#else
          t=libnet_build_icmpv4_redirect(ICMPType, ICMPCode, 0, DestIP, 
                                         IPlen, 0, IPid, IPFrag, IPttl, 
                                         IPproto, 0, DestIP, SrcIP, Payload, 
                                         PayloadSize, l ,0);
#endif
	  if (t == -1)
	    {
	      fprintf(stderr, "Can't build ICMP header: %s\n", 
		      libnet_geterror(l));
	      libnet_destroy(l);
	      return (EXIT_FAILURE);
	    }
	  t = libnet_build_ipv4(LIBNET_IPV4_H + LIBNET_ICMPV4_REDIRECT_H 
				+ PayloadSize, 0, IPid, IPFrag, IPttl,
				IPPROTO_ICMP, 0, SrcIP, DestIP, NULL, 0, l, 0);
	  if (t == -1)
	    {
	      fprintf(stderr, "Can't build IP header: %s\n", libnet_geterror(l));
	      libnet_destroy(l);
	      return (EXIT_FAILURE);
	    }
	  HavePacket=1;
	  break;
       
	             /* time exceed */
	case 11: 
	  IPlen=GetRandom(65535);

#ifdef HAVE_LIBNET_LIBNET_TYPES_H
	  t=libnet_build_icmpv4_timeexceed (ICMPType, ICMPCode, 0,
					    Payload, 
					    PayloadSize, l ,0);
#else
          t=libnet_build_icmpv4_timeexceed (ICMPType, ICMPCode, 0, IPlen,
                                            0, IPid, IPFrag, IPttl, IPproto,
                                            0, SrcIP, DestIP, Payload, 
                                            PayloadSize, l ,0);
#endif
	  if (t == -1)
	    {
	      fprintf(stderr, "Can't build ICMP header: %s\n", 
		      libnet_geterror(l));
	      libnet_destroy(l);
	      return (EXIT_FAILURE);
	    }
	  t = libnet_build_ipv4(LIBNET_IPV4_H + LIBNET_ICMPV4_TIMXCEED_H 
				+ PayloadSize, 0, IPid, IPFrag, IPttl,
				IPPROTO_ICMP, 0, SrcIP, DestIP, NULL, 0, l, 0);
	  if (t == -1)
	    {
	      fprintf(stderr, "Can't build IP header: %s\n", libnet_geterror(l));
	      libnet_destroy(l);
	      return (EXIT_FAILURE);
	    }
	  HavePacket=1;

                       /* timestamp */
	case 13:
	case 14: 
	  /* we misuse DestIP and SrcIP as time values...*/

	  t=libnet_build_icmpv4_timestamp(ICMPType, ICMPCode, 0, ICMPId, 
					  ICMPSeq, DestIP, SrcIP, DestIP, 
					  Payload, PayloadSize, l ,0);
	  if (t == -1)
	    {
	      fprintf(stderr, "Can't build ICMP header: %s\n", 
		      libnet_geterror(l));
	      libnet_destroy(l);
	      return (EXIT_FAILURE);
	    }

	  t = libnet_build_ipv4(LIBNET_IPV4_H + LIBNET_ICMPV4_TS_H 
				+ PayloadSize, 0, IPid, IPFrag, IPttl,
				IPPROTO_ICMP, 0, SrcIP, DestIP, NULL, 0, l, 0);
	  if (t == -1)
	    {
	      fprintf(stderr, "Can't build IP header: %s\n", libnet_geterror(l));
	      libnet_destroy(l);
	      return (EXIT_FAILURE);
	    }
	  HavePacket=1;

	case 17:
	case 18: /* mask */
	  t=libnet_build_icmpv4_mask (ICMPType, ICMPCode, 0, ICMPId, ICMPSeq,
				      DestIP, Payload, PayloadSize, l ,0);
	  if (t == -1)
	    {
	      fprintf(stderr, "Can't build ICMP header: %s\n", 
		      libnet_geterror(l));
	      libnet_destroy(l);
	      return (EXIT_FAILURE);
	    }

	  t = libnet_build_ipv4(LIBNET_IPV4_H + LIBNET_ICMPV4_MASK_H 
				+ PayloadSize, 0, IPid, IPFrag, IPttl,
				IPPROTO_ICMP, 0, SrcIP, DestIP, NULL, 0, l, 0);
	  if (t == -1)
	    {
	      fprintf(stderr, "Can't build IP header: %s\n", libnet_geterror(l));
	      libnet_destroy(l);
	      return (EXIT_FAILURE);
	    }
	  HavePacket=1;
	  break;
	  
	default:
	  printf("Not supported ICMP - Type: %u, Code: %u\n", 
		 ICMPType, ICMPCode);
	  HavePacket=0;
	}
      if (HavePacket) ICMPCount++;      
    }
  
  if(!HavePacket) 
    {
       /* We have no packet so we ignore this rule and return */
      IgnoredRules++;
      libnet_destroy(l);
      return(1);
    }

  /*
   *  Write it to the wire.
   */
  for (i=0; i<RepeatCycles; i++)
    {
      c = libnet_write(l);
      if (c == -1)
	{
	  FailedCount++;
	  fprintf(stderr, "Write error: %s", libnet_geterror(l));
	  libnet_destroy(l);
	  return (EXIT_FAILURE);
	}
      else
	{
	  TotalBytes += c;
	  TotalPackets++;
#ifdef DEBUG
          printf("TotalPackets: %u (%u)\n",TotalPackets,MaxPackets);
#endif
          /* Did we reach maximum of packets to send? */
          if( (MaxPackets) && (TotalPackets==MaxPackets))
             {
                libnet_destroy(l);
                return(EXIT_SUCCESS);
             }
	}
      if ((Delay) && (i%Delay)==0) usleep(TDelay);
      if (Random) usleep(GetRandom(Random));
    }      

  libnet_destroy(l);
  return (EXIT_SUCCESS);
}

void 
SigInt()
{
  DumpStatistics();
  exit(0);
}

void
SetIsDataAt(char *value)
{
  ContentList *List;
  char *key=NULL;
  int relative=0;
  int offset=0;

  if ( (key=strchr(value,',')) )
   {
    *key='\0';
     key++; 

#ifdef DEBUG
     printf("Key: <%s>\n",key);
#endif

     if (strncmp(key,"relative",8))
     {
        printf("Invalid key: <%s> -> Ingoring\n",key);
        return;
     }
    relative=1; 
   }

   offset=atoi(value);

#ifdef DEBUG
   printf("Value  : <%s>\n",value);
   printf("Offset1: <%i>\n",offset);
#endif

   if (relative)
   {
      List=StartContent;
      while ( List->next != NULL )
       {
         offset+= List->offset + List->length;
         List = List->next; 
       }
   }

   if (mlength < offset) mlength=offset;

#ifdef DEBUG
   printf("Offset2: <%i>\n",offset);
   printf("mlength: <%i>\n",mlength);
#endif
   return;
}
