#include "comm.h"
#include "io.h"
#include "FLoPconfig.h"
#include "flopdebug.h"


extern u_int LogSyslog; /* defined in main program */

#ifdef DEBUGFLOP
extern u_int8_t DebugVector[DEBUG_MAX]; /* defined in main program */
#endif


/*
 * Function: Printf(format,args,...)
 * 
 * Purpose: prints either to stdout or log via syslog depending
 *          on global LogSyslog flag
 *
 * Author: DG
 *
 */

void 
Printf(char *format,...)
{
  char line[STRLEN]="";
  va_list va_args;

  bzero(line,STRLEN);
  va_start(va_args,format);
  vsnprintf(line,STRLEN-1,format,va_args);
  va_end(va_args);

  if (LogSyslog)
    syslog(LOG_INFO|LOG_LOCAL0,line);
  else
    printf(line);
}


/*
 * Function: DG_strerror_r(errnum,buf,n)
 * 
 * Purpose: Our own version of strerror_r, there are several 
 *          incompatible verions out there, especially glibc 
 *          which returns a char pointer
 *
 * Arguments: error number, buffer to store message, size of buffer
 *
 * Author: DG
 *
 */

int 
DG_strerror_r(int errnum, char *buf, size_t n)
{
   return snprintf(buf, n, strerror(errnum));
}


/*
 * Function: CheckDirectory
 * 
 * Purpose: Check accessibility of directory to create swap
 *          files for the case the database dies or fails.
 *          We need all rights, read, write and execute to
 *          change in the directory.
 *
 * Arguments: string containing the directory
 *
 * Return:  0 if all is ok
 *         -1 on error
 *
 * Author: DG
 *
 */
int 
CheckDirectory(char *dir)
{
  char errbuf[STRLEN];

  /* First we check the directory */
  if  (access(dir, R_OK|W_OK|X_OK) <0 )
      {
	DG_strerror_r(errno,errbuf,STRLEN);
	Printf("Acess error on <%s>: %s\n",dir,errbuf);
         return(-1);
       }
    else
      {
	DEBUGF(DebugOutput(DEBUG_RD,3,"CheckDirectory: Access ok.\n"););
	return(0);
       }
}


/*
 * Function: CheckFile
 * 
 * Purpose: Check if a swap file is present and if the size
 *          is greater zero -> we have to read in swapped data
 *
 * Arguments: string containing the file
 *
 * Return:  0 if nothing is to do
 *         -1 on error
 *          1 file is present and not empty
 *
 * Author: DG
 *
 */
int 
CheckFile(char *file)
{
  char errbuf[STRLEN];
  struct stat buf;

   /* Check if file exits, is accessable and get the size */
    if (stat(file,&buf))
     {
       /* File does not exist */
        DG_strerror_r(errno,errbuf,STRLEN);
        Printf("CheckFile -> %s: %s\n",file, errbuf);
	return(0);
     }
    else
     {
       /* File exists, test if we have rw access */
        if (access(file, R_OK|W_OK) < 0 )
	 {
	   DG_strerror_r(errno,errbuf,STRLEN);
	   Printf("CheckFile, can not access file -> %s: %s\n",file, errbuf);
	   return(-1);
         }
	else
	 {

	DEBUGF(DebugOutput(DEBUG_RD,3,
			   "CheckFile> Access on file %s is ok.\n",file););
	DEBUGF(DebugOutput(DEBUG_RD,3,
			   "CheckFile> Size: %i\n",(int) buf.st_size););

           if (buf.st_size>0) return(1);
         }
     }
    return(0);
}


/*
 * Function: DGrecv(socket, buffer, length, flags)
 * 
 * Purpose: wrapper program around recv() system call to
 *          avoid interrupted system calls due to signals
 *
 *          This is needed on all BSD systems although
 *          SA_RESTART is set in the signalhandler.
 *
 * Returns: on error -> -1
 *          else     -> number of received bytes
 *
 * Author: DG
 *
 */

int 
DGrecv(int sock, void *buf, size_t len, int flags)
{
  int tbrecv=0,brecv=0;
 
  while (len > tbrecv )
    {
      brecv=recv(sock, buf+tbrecv, len-tbrecv, flags);

      switch (brecv) 
	{
	case -1: 
	  if (errno == EINTR || errno == EAGAIN) 
	    {
	      continue;
	    }
	  else 
	    {
              Printf("DGrecv error -> flags: %i errno: %i - total read: %i,"
                     " actual:%i)\n", flags,errno,tbrecv,brecv);
	      return(-1);
	    }
	  break;
	  
	case 0:
	  return(tbrecv);
	  break;
 
	default:
	  tbrecv += brecv;
	}
    }

  return(tbrecv);
  
}


/*
 * Function: DGsend(socket, buffer, length, flags)
 * 
 * Purpose: wrapper program around send() system call to
 *          avoid interrupted system calls due to signals
 *
 *          This is needed on all BSD systems although
 *          SA_RESTART is set in the signalhandler.
 *
 * Returns: on error -> -1
 *          else     -> number of received bytes
 *
 * Author: DG
 *
 */

int 
DGsend(int sock, void *buf, size_t len, int flags)
{
  int tbsend=0,bsend=0;

  while (len > tbsend )
    {
      bsend=send(sock,buf+tbsend,len-tbsend,flags);
      switch (bsend) 
	{
	case -1: 
	  if (errno == EINTR || errno == EAGAIN) 
	    {
	      continue;
	    }
	  else 
	    {
	      return(-1);
	    }
	  break;
	  
	case 0:
	  return(tbsend);
	  break;
 
	default:
	  tbsend += bsend;
	}
    }
  
  return(tbsend);
}

void 
set_nonblock(int fd)
{
        int val;

        val = fcntl(fd, F_GETFL, 0);
        if (val < 0) 
          {
                Printf("fcntl(%d, F_GETFL, 0): %s", fd, strerror(errno));
                return;
          }
        if (val & O_NONBLOCK) 
          {
	    DEBUGF(DebugOutput(DEBUG_UTIL,5,
			       "fd %d is O_NONBLOCK\n", fd););
             return;
           }

	DEBUGF(DebugOutput(DEBUG_UTIL,5,"fd %d setting O_NONBLOCK\n", fd););

        val |= O_NONBLOCK;
        if (fcntl(fd, F_SETFL, val) == -1)
                Printf("fcntl(%d, F_SETFL, O_NONBLOCK): %s",
                    fd, strerror(errno));
}

#ifdef DEBUGFLOP
/*
 * Function: DebugOutput(format,args,...)
 * 
 * Purpose: prints debug output to stderr or via syslog depending
 *          on global LogSyslog flag
 *
 * Author: DG
 *
 */

void 
DebugOutput(u_int8_t area, u_int8_t level, char *format,...)
{
  char line[STRLEN]="";
  va_list va_args;

  if (DebugVector[area] == 0) return;
  if (level > DebugVector[area]) return;
  bzero(line,STRLEN);
  snprintf(line,13,"DEBUG(%2u:%1u) ",area,level);
  va_start(va_args,format);
  vsnprintf(line+12,STRLEN-13,format,va_args);
  va_end(va_args);

  if (LogSyslog)
    syslog(LOG_INFO|LOG_LOCAL0,line);
  else
    fprintf(stderr,line);
}

#endif
