/*
** Some parts are taken from the snort project and were modified, so
**
** Partial Copyright (C) 1998-2002 Martin Roesch <roesch@sourcefire.com>
** Partial Copyright (C) 2002-2004 Dirk Geschke <Dirk@geschke-online.de>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "socket_database.h"
#include "flopdebug.h"
#include <time.h>

extern u_int UTC;

static char * time_ptr=NULL;
static char * retbuf = NULL;

/* in util.c: we first try to allocate as much memory as */
/* possible on startup. Thus we omit later checking for */
/* free memory */
void 
util_alloc()
{
  DEBUGF(DebugOutput(DEBUG_UTIL,5,"Util_alloc> Start to allocate memory\n"););

    /* allocate only memory the first time */
    if (time_ptr==NULL)
      {
	time_ptr = (char *)malloc(SMALLBUFFER);
	if (time_ptr == NULL)
	  {
	    Printf("GetTimeStamp: problem to allocte memory.\n");
	    Printf("Exiting.\n");
	    exit(1);
	  }
      }

    if (retbuf == NULL)
      {
	 retbuf = (char *) malloc((MAXBUFFSIZE*2)+1);
	 if ( retbuf == NULL )
	   {
	     Printf("Base64: problem to allocate memory -> Exiting.\n");
	     Printf("Exiting.\n");
	     exit(1);
	   }
      }

}

int 
GetLocalTimeZone()
{
    struct timezone tz;
    struct timeval tv;

    gettimeofday(&tv,&tz);

    if(tz.tz_minuteswest > 720) return(24 - tz.tz_minuteswest/60);
    else          return(0 - tz.tz_minuteswest/60);
}


/****************************************************************************
 *
 * Function: GetTimeStamp(time_t * tv_sec, time_t * tv_usec, int tz)
 *
 * Purpose: Get an ISO-8601 formatted timestamp for tv_sec within the tz
 *          timezone.
 *
 * Arguments: tv_sec is a time_t pointer. tz is a timezone.
 *
 * Returns: char * -- You must free this char * when you are done with it.
 *
 ***************************************************************************/
char *
GetTimeStamp(struct timeval * ts, int tz)
{
    struct tm *time;
    int sec;
    long usec;

    /* allocate only memory the first time */
    if (time_ptr==NULL)
      {
	time_ptr = (char *)malloc(SMALLBUFFER);
	if (time_ptr == NULL)
	  {
	    Printf("GetTimeStamp: problem to allocte memory.\n");
	    Printf("Exiting.\n");
	    exit(1);
	  }
      }

    bzero(time_ptr,SMALLBUFFER);
    sec =(int)&ts->tv_sec;
    usec=(long)ts->tv_usec;
    if(UTC)
        time = gmtime((time_t *)&ts->tv_sec);
    else
        time = localtime((time_t *)&ts->tv_sec);

    if(tz < 0)
    {
        snprintf(time_ptr, SMALLBUFFER, 
		 "%04i-%02i-%02i %02i:%02i:%02i.%06li%03i",
                 1900 + time->tm_year, time->tm_mon + 1, time->tm_mday,
                 time->tm_hour, time->tm_min, time->tm_sec, usec, tz);
    }
    else
    {
        snprintf(time_ptr, SMALLBUFFER, 
		 "%04i-%02i-%02i %02i:%02i:%02i.%06li+%02i",
                 1900 + time->tm_year, time->tm_mon + 1, time->tm_mday,
                 time->tm_hour, time->tm_min, time->tm_sec, usec, tz);
    }

    return time_ptr;
}


/****************************************************************************
 *
 * Function: GetCurrentTimeStamp()
 *
 * Purpose: Generate an ISO-8601 formatted timestamp for the current time.
 *
 * Arguments: none
 *
 * Returns: char * -- You must free this char * when you are done with it.
 *
 ***************************************************************************/
char *
GetCurrentTimeStamp()
{
    struct tm *lt;
    struct timezone tz;
    struct timeval tv;
    struct timeval *tvp;
    int tzone;

    /* allocate only memory the first time */
    if (time_ptr==NULL)
      {
	time_ptr = (char *)malloc(SMALLBUFFER);
	if (time_ptr == NULL)
	  {
	    Printf("GetCurrentTimeStamp: problem to allocate memory.\n");
	    Printf("Exiting.\n");
	    exit(1);
	  }
      }

    bzero(time_ptr,SMALLBUFFER);
    bzero((char *)&tz,sizeof(tz));
    gettimeofday(&tv,&tz);
    tvp = &tv;

    if(UTC)
    {
        lt = gmtime((time_t *)&tvp->tv_sec);
        snprintf(time_ptr, SMALLBUFFER, "%04i-%02i-%02i %02i:%02i:%02i",
                1900 + lt->tm_year, lt->tm_mon + 1, lt->tm_mday,
                lt->tm_hour, lt->tm_min, lt->tm_sec);
    }
    else
    {
        lt = localtime((time_t *)&tvp->tv_sec);

        tzone = GetLocalTimeZone();

        if(tzone <= 0)
            snprintf(time_ptr, SMALLBUFFER,
                    "%04i-%02i-%02i %02i:%02i:%02i%03i",
                    1900 + lt->tm_year, lt->tm_mon + 1, lt->tm_mday,
                    lt->tm_hour, lt->tm_min, lt->tm_sec, tzone);
        else
            snprintf(time_ptr, SMALLBUFFER,
                    "%04i-%02i-%02i %02i:%02i:%02i+%02i",
                    1900 + lt->tm_year, lt->tm_mon + 1, lt->tm_mday,
                    lt->tm_hour, lt->tm_min, lt->tm_sec, tzone);
    }

    return time_ptr;
}

/****************************************************************************
 *
 * Function: Ascii(u_char *xdata, int length)
 *
 * Purpose: This function takes takes a buffer "xdata" and its length then
 *          returns a string of only the printable ASCII characters.
 *
 * Arguments: xdata is the buffer, length is the length of the buffer in
 *            bytes
 *
 * Returns: char * -- You must free this char * when you are done with it.
 *
 ***************************************************************************/
char *
Ascii(u_char *xdata, int length)
{
    int x;
    int y;
    char *buf;

    if (retbuf == NULL)
      {
	 retbuf = (char *) malloc((MAXBUFFSIZE*2)+1);
	 if ( retbuf == NULL )
	   {
	     Printf("Base64: problem to allocate memory -> Exiting.\n");
	     Printf("Exiting.\n");
	     exit(1);
	   }
      }

    buf=retbuf;

    y = 0;

    for(x = 0; ((x < length) && (x + y + 4 < length * 1.1)); x++)
    {
        if((xdata[x] > 0x1F) && (xdata[x] < 0x7F))
        {
            if(xdata[x] == '<')
            {
                *buf = '&'; buf++; y++;
                *buf = 'l'; buf++; y++;
                *buf = 't'; buf++; y++;
                *buf = ';';
            }
            else if(xdata[x] == '&')
            {
                *buf = '&'; buf++; y++;
                *buf = 'a'; buf++; y++;
                *buf = 'm'; buf++; y++;
                *buf = 'p'; buf++; y++;
                *buf = ';';
            }
            else if(xdata[x] == '>')
            {
                *buf = '&'; buf++; y++;
                *buf = 'g'; buf++; y++;
                *buf = 't'; buf++; y++;
                *buf = ';';
            }
            else if(xdata[x] == '\'')
            {
                *buf = '\\'; buf++; y++;
                *buf = '\'';
            }
            else
            {
                *buf = xdata[x];
            }
            buf++;
        }
        else
        {
            *buf = '.'; buf++;
        }
    }

    *buf = '\0';
    return retbuf;
}

char *
Fasthex(u_char *xdata, int length)
{
    char conv[] = "0123456789ABCDEF";
    char *index;
    char *end;
    char *ridx;

    index = (char *) xdata;
    end = index + length;

    DEBUGF(DebugOutput(DEBUG_UTIL,5,"Fasthex> retbuf %p\n",retbuf););
    DEBUGF(DebugOutput(DEBUG_UTIL,5,"Fasthex> xdata at %p, len: %u\n",
		       xdata,length););

    if (retbuf == NULL)
      {
	 retbuf = (char *) malloc((MAXBUFFSIZE*2)+1);
	 if ( retbuf == NULL )
	   {
	     Printf("Fasthex: Problem to allocate memory.\n");
	     Printf("Exiting.\n");
	     exit(1);
	   }
      }
    ridx = retbuf;

    while(index < end)
    {
        *ridx++ = conv[((*index & 0xFF)>>4)];
        *ridx++ = conv[((*index & 0xFF)&0x0F)];
        index++;
    }

    *ridx='\0';

    return retbuf;
}

/*
 * Function: Base64(char * xdata, int length)
 *
 * Purpose: Insert data into the database
 *
 * Arguments: xdata  => pointer to data to base64 encode
 *            length => how much data to encode
 *
 * Make sure you allocate memory for the output before you pass
 * the output pointer into this function. You should allocate
 * (1.5 * length) bytes to be safe.
 *
 * Returns: data base64 encoded as a char *
 *
 */
char * 
Base64(u_char * xdata, int length)
{
    int count, cols, bits, c, char_count;
    unsigned char alpha[64] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
    char * output;
    char_count = 0;
    bits = 0;
    cols = 0;

    if (retbuf == NULL)
      {
	 retbuf = (char *) malloc((MAXBUFFSIZE*2)+1);
	 if ( retbuf == NULL )
	   {
	     Printf("Base64: problem to allocate memory -> Exiting.\n");
	     Printf("Exiting.\n");
	     exit(1);
	   }
      }

    output = retbuf;

    for(count = 0; count < length; count++)
    {
        c = xdata[count];

        if(c > 255)
        {
            Printf("Base64(): encountered char > 255 (decimal%d)\n If you see this error message a char is more than one byte on your machine\n This means your base64 results can not be trusted", c);
        }

        bits += c;
        char_count++;
        if(char_count == 3)
        {
            *output = alpha[bits >> 18]; output++;
            *output = alpha[(bits >> 12) & 0x3f]; output++;
            *output = alpha[(bits >> 6) & 0x3f]; output++;
            *output = alpha[bits & 0x3f]; output++;
            cols += 4;
            if(cols == 72)
            {
                *output = '\n'; output++;
                cols = 0;
            }
            bits = 0;
            char_count = 0;
        }
        else
        {
            bits <<= 8;
        }
    }

    if(char_count != 0)
    {
        bits <<= 16 - (8 * char_count);
        *output = alpha[bits >> 18]; output++;
        *output = alpha[(bits >> 12) & 0x3f]; output++;
        if(char_count == 1)
        {
            *output = '='; output++;
            *output = '='; output++;
        }
        else
        {
            *output = alpha[(bits >> 6) & 0x3f];
            output++; *output = '=';
            output++;
        }
    }
    *output = '\0';
    return retbuf;
}


